/*
 * Implements OMP2D in c++, only accepts real arguments,
 * stops for a fixed tol, must have 3 inputs and five 
 * outputs.
 * MATLAB CALL
 *  [H,Di1,Di2, beta, c, Q, DiXY] = OMP2D( I, Dx, Dy, tol, No, indx, indy);
 */
#define output 0
#include "mex.h"
#include "../../Common/MatrixOperations.h"
#include "OMP2D.h"
using namespace std;

void mexFunction(int nlhs, mxArray *plhs[],
    int nrhs, const mxArray *prhs[])
{        
    const char name[] = "OMP2DCPP";          
            
    /* Begin input validation */    
    if((nrhs < 3) || (nrhs > 8))
    {
        mexErrMsgTxt("Wrong number of input arguments");
    }
    
    // Check we are being passed real full and non string variables     
    for ( int i = 0; i < nrhs; i++ )
    {        
        if ( mxIsComplex(prhs[i]) || mxIsClass(prhs[i],"sparse") || mxIsChar(prhs[i]) || !mxIsDouble(prhs[i]) )
        {
            mexErrMsgTxt("Input must be double, real, full,and nonstring");
        }
    }
    
    /* Validate input dimensions */    
    /* Check mImage is not empty */
    if (mxIsEmpty(prhs[0]))
    {
        mexErrMsgTxt("The image matrix is empty!!\n");
    }
    
    /* mImage should be a matrix but routine should run as long as size(I,1) == size(mDictionaryY,1) and size(I,2) ==  size(mDictionaryX,1)*/
    unsigned long szSpaceY = mxGetM(prhs[0]);
    unsigned long szSpaceX = mxGetN(prhs[0]);
    unsigned long szSpaceYX = szSpaceY*szSpaceX;
        
    /* Validate mImage against the sizes of mDictionaryY and mDictionaryX */
    if ((mxGetM(prhs[1]) != szSpaceY) || mxIsEmpty(prhs[1]))
    {
        mexPrintf("The atoms in the dictionary dictionary Y are of length %i and should be of length %i\n",
                mxGetM(prhs[1]),szSpaceY);
        mexErrMsgTxt("Error!!\n");
    }
    if ((mxGetM(prhs[2]) != szSpaceX) || mxIsEmpty(prhs[2]))
    {
        mexPrintf("The atoms in the dictionary dictionary X are of length %i and should be of length %i\n",
                mxGetM(prhs[2]),szSpaceX);
        mexErrMsgTxt("Error!!\n");
    }
    
    /* Get nAtomsY and nAtomsX check at least one of them is greater than 1 and the other is at least 1*/
    // As they are not empty if the sum is greater than 2 the above should be satisfied.
    unsigned long nAtomsY = mxGetN(prhs[1]);
    unsigned long nAtomsX = mxGetN(prhs[2]);
    unsigned long nAtomsYX = nAtomsY*nAtomsX;
    if ( (nAtomsX + nAtomsY) < 3 )
    {
        mexPrintf("Dictionary Y has %i atoms and dictionary X has %i atoms\n!",nAtomsY,nAtomsX);
        mexErrMsgTxt("One dictionary should have more than 1 atom and the other should have at least one!!");
    }
            
    /* Check tolerance is a scalar greater than 1 */
    double tolerance1 = 6.5;
    if(nrhs >= 4){
        if (!mxIsEmpty(prhs[3]))
        {
            if (( mxGetM(prhs[3]) != 1) || ( mxGetN(prhs[3]) != 1))
            {
                mexErrMsgTxt("The tolerance should be a scalar");
            }
            tolerance1 = *mxGetPr(prhs[3]);
        }
    }
    
    /* Get maxDimensions for big arrays as th min(szSpaceX*szSpaceY and nAtomsX*nAtomsY) call it maxIterations */
    unsigned long maxDimensions = min(szSpaceYX,nAtomsYX);
      
    /* Get No and validate if not an input set to zero */
    unsigned long No  = 0;
    if(nrhs >= 5){
        if (!mxIsEmpty(prhs[4]))
        {
            if (( mxGetM(prhs[4]) != 1) || ( mxGetN(prhs[4]) != 1))
            {
                mexErrMsgTxt("The maximum number of atoms should be a scalar");
            }
            No = static_cast<unsigned long>(*mxGetPr(prhs[4]));
        }
    }
    
    /* Validate indY and indX to make sure they are 1D vectors not exceeding the size of the space */
    /* Make sure both are present */
    if (nrhs == 6) 
    {
        mexErrMsgTxt("Index's from dictionaries X and Y for the initial subspace need to be provided!!");
    }
    
    
    
    // First indX
    unsigned long numIndX = 0;
    double *indX = 0;
    if(nrhs >= 6){
        if (!mxIsEmpty(prhs[5]))
        {
            if (( mxGetM(prhs[5]) == 1) && ( mxGetN(prhs[5]) >= 0))
            {
                /* Check largest dimension does not exceed the size of the space */
                if( mxGetN(prhs[5]) > szSpaceYX)
                {
                    mexErrMsgTxt("The initial subspace is too large");
                }else{
                    numIndX = mxGetN(prhs[5]);
                }
            }else if(( mxGetM(prhs[5]) >= 0) && ( mxGetN(prhs[5]) == 1))
            {
                /* Check largest dimension does not exceed the size of the space */
                if( mxGetM(prhs[5]) >szSpaceYX)
                {
                    mexErrMsgTxt("The initial subspace is too large");
                }else{
                    numIndX = mxGetM(prhs[5]);
                }
            }else{
                mexErrMsgTxt("The index's of the initial subspace should be a vector");
            }
            indX = mxGetPr(prhs[5]);
        }
    }
    
    
    // Next indY additionaly validate same numner of elements as indY
    double *indY = 0;
    if(nrhs >= 7){
        if (!mxIsEmpty(prhs[6]))
        {
            if (( mxGetM(prhs[6]) == 1) && ( mxGetN(prhs[6]) >= 0))
            {
                /* Check largest dimension does not exceed the size of the space */
                if( mxGetN(prhs[6]) > szSpaceYX)
                {
                    mexErrMsgTxt("The initial subspace is too large");
                }else if( mxGetN(prhs[6]) != numIndX)
                { 
                    mexErrMsgTxt("The index of the initial subspace for X does not agree with Y!");
                }
            }else if(( mxGetM(prhs[6]) >= 0) && ( mxGetN(prhs[6]) == 1))
            {
                /* Check largest dimension does not exceed the size of the space */
                if( mxGetM(prhs[6]) >szSpaceYX)
                {
                    mexErrMsgTxt("The initial subspace is too large");
                }else if( mxGetM(prhs[6]) != numIndX)
                { 
                    mexErrMsgTxt("The index of the initial subspace for X does not agree with Y!");
                }
            }else{
                mexErrMsgTxt("The index's of the initial subspace should be a vector");
            }
            indY = mxGetPr(prhs[6]);
        }else{
            /* Make sure that indX was also empty (numIndY == 0) */
            if (numIndX !=0)
            {
                mexErrMsgTxt("Initial index's only supplied for Y!!");
            }
        }
    }
        
    /* Begin Output validation */
    if ( nlhs < 6)
    {
        mexErrMsgTxt("Wrong number of output arguments!");
    }

    /* Get pointers to the remaining input data */    
    double *mImage = mxGetPr(prhs[0]);
    double *mDictionaryY = mxGetPr(prhs[1]);
    double *mDictionaryX = mxGetPr(prhs[2]);    
 
    /* Create mxArray's for the output data */
    // Approximated image matrix size szSpaceY x szSpaceX
    plhs[0] = mxCreateDoubleMatrix(szSpaceY, szSpaceX, mxREAL);
    // Index's in mDictionaryY allocate after calling OMP2D as depends on nChosen
    // Index's in mDictionaryX allocate after calling OMP2D as depends on nChosen
    // Biorthogonal atoms allocate after calling OMP2D as depends on nChosen
    // Coefficients of decomoposition after calling OMP2D as depends on nChosen
    // Orthogonal atoms allocate after calling OMP2D as depends on nChosen
 	// Index's in mDictionaryXY allocate after calling OMP2D as depends on nChosen OPTIONAL

    /* Create pointer's to the output data */
    double *mApproxImage = mxGetPr(plhs[0]);
        
    // Set arrays to the max possible size for the output arguments that depent on nChosen
    double *vIYBig = new double[maxDimensions];
    double *vIXBig = new double[maxDimensions];    	  
    double *mBetaBig = new double[szSpaceYX * maxDimensions];
    double *vCoefficientsBig = new double[maxDimensions];            
    double *mOrthogonalBig = new double[szSpaceYX * maxDimensions];
    double *vIXYBig = 0;
	bool bigIndex = 0;
	if (nlhs == 7 )
	{
		vIXYBig = new double[maxDimensions];
		bigIndex = 1;      
	}

    unsigned long nChosen = 0;
    
    int fail = OMP2D(mApproxImage, mImage, mDictionaryY, mDictionaryX, tolerance1, szSpaceX, szSpaceY, nAtomsX, nAtomsY,
        vIYBig, vIXBig, vIXYBig, bigIndex, mBetaBig,vCoefficientsBig, mOrthogonalBig, No, indY, indX, numIndX, nChosen);
    
    /* Check if I can just copy without declaring pointers */
    if(~fail && nChosen <= maxDimensions)
    {
        /* Copy the oversized arrays to smaller arrays for output now we know nChosen */
        // Index's in mDictionaryY size nChosen
        plhs[1] = mxCreateDoubleMatrix(1,nChosen, mxREAL);
        //double *vIY = mxGetPr(plhs[1]);
        CopyElements(vIYBig,mxGetPr(plhs[1]),nChosen);



        // Index's in mDictionaryX size nChosen
        plhs[2] = mxCreateDoubleMatrix(1,nChosen, mxREAL);   
        //double *vIX = mxGetPr(plhs[2]);
        CopyElements(vIXBig,mxGetPr(plhs[2]),nChosen);
        // Biorthogonal atoms size szSpaceYX*nChosen
        plhs[3] = mxCreateDoubleMatrix(szSpaceYX, nChosen, mxREAL);
        //double *mBiorthogonal = mxGetPr(plhs[3]);
        CopyElements(mBetaBig,mxGetPr(plhs[3]),szSpaceYX*nChosen);
        // Coefficients of decomoposition size nChosen                
        plhs[4] = mxCreateDoubleMatrix(1, nChosen, mxREAL);
        //double *vCoefficients = mxGetPr(plhs[4]);
        CopyElements(vCoefficientsBig,mxGetPr(plhs[4]),nChosen);
        // Orthogonal atoms size szSpaceYX*nChosen
        plhs[5] = mxCreateDoubleMatrix( szSpaceYX,nChosen, mxREAL);
        //double *mOrthogonal = mxGetPr(plhs[5]);
        CopyElements(mOrthogonalBig,mxGetPr(plhs[5]),szSpaceYX*nChosen); 
		if (bigIndex)
		{
			// Index's in mDictionaryX size nChosen
			plhs[6] = mxCreateDoubleMatrix(1,nChosen, mxREAL);   
			//double *vIXY = mxGetPr(plhs[6]);
			CopyElements(vIXYBig,mxGetPr(plhs[6]),nChosen);
		}
    }
    
    /* Delete dynamic arrays to free memory */
    delete[] vIYBig;
    delete[] vIXBig;
    delete[] mBetaBig;
    delete[] vCoefficientsBig;
    delete[] mOrthogonalBig;
    vIYBig = 0;
    vIXBig = 0;
    vCoefficientsBig = 0;   
    mBetaBig = 0;
    mOrthogonalBig = 0;  
    delete [] vIXYBig;
    vIXYBig = 0;
	
        
            
    /* Check if OMP2D has failed and that nChosen does not exceed max iterations */
    if(fail || nChosen > maxDimensions)
    {
        mexErrMsgTxt("Error OMP2D exited early see message above!!\n");
    }
}
