/*
 * Gateway to implementation of ProjMP2D in c++
 *
 * 2DProjMP Projected Matching Pursuit 
 *
 * It uses the MP method for projecting the MP solution onto the spann of 
 * selected %2D atoms
 *
 * Usage:    [h,Re, c] = ProjMP2D(h,Re V1,V2,c,toln,Max);
 * [h,Re, c] = ProjMP2D(h,Re V1,V2,c);
 *
 *
 * Inputs:
 * Re     Residual of MP
 * h      Approximation by MP
 * Re     Residual of MP
 * V1     1D Dictionary selected by MP (horizontal coordinate)
 * V2     1D Dictionary selected by MP (vertical coordinate)
 * c      coefficientes in the MP approximation
 * toln   numerical tolerance for the projection  (default tre=1e-3)
 * Max    maximum number of iterations (default Max=3000)
 *
 * Outputs:
 * h     projected approximation  
 * Re    projected residule
 * c     coefficients in the decomposition
 */

#define output 0
#include "mex.h"
#include "ProjMP2D.cpp"
using namespace std;

/* Function Declarations */
static void ProjMP2D_Mex( int nlhs, mxArray *plhs[], int nrhs, const 
        mxArray *prhs[]);
double *GetResidule(const mxArray *prhs,unsigned long &nRows,
        unsigned long &nColumns);
double *GetApprox(const mxArray *prhs, ulong nRowsResidule,
        unsigned long nColumnsResidule);
void ValidateApprox(unsigned long nRowsResidule, unsigned long 
        nColumnsResidule, unsigned long nRows, unsigned long nColumns);
double *GetDictColumns(const mxArray *prhs, unsigned long nRowsResidule,
        unsigned long &nAtomsColumn);     
void ValidateDictColumn(unsigned long nRowsResidule, unsigned long nRows);
double *GetDictRows(const mxArray *prhs,unsigned long nColumnsResidule,
        unsigned long nAtomsColumn);
void ValidateDictRow(unsigned long nColumnsResidule, unsigned long 
        nAtomsColumn, unsigned long nRows, unsigned long nColumns);
double *GetCoefficients(const mxArray *prhs, unsigned long nAtoms);
void ValidateCoefficients(unsigned long nAtoms, unsigned long nRows,
        unsigned long nColumns);
double GetTolerance(const mxArray *prhs);
unsigned long GetMaxIterations(const mxArray *prhs);
void ValidateScalar(unsigned long nRows, unsigned long nColumns, const char 
        *varName);
double *GetVar(const mxArray *prhs, unsigned long &M, unsigned long &N, 
        const char *varName);
double *SetVar(mxArray *&m, unsigned long nRows, unsigned long 
        nColumns);

void mexFunction(int nlhs, mxArray *plhs[],
    int nrhs, const mxArray *prhs[])
{
    /* Gateway from MATLAB */
    ProjMP2D_Mex(nlhs,plhs,nrhs,prhs);
}

static void ProjMP2D_Mex( int nlhs, mxArray *plhs[],
        int nrhs, const mxArray *prhs[])
{
    /* Validate input to function */
        
    /* Validate number of input (rhs) arguments*/    
    if((nrhs < 5) || (nrhs > 7))
    {
        mexErrMsgTxt("Wrong number of input arguments");
    }
    
    /* Check we are being passed real full and double variables */
    for ( int i = 0; i < nrhs; i++ )
    {        
        if ( mxIsComplex(prhs[i]) || mxIsClass(prhs[i],"sparse")|| 
                !mxIsDouble(prhs[i]) )
        {
            mexErrMsgTxt("Input must be double, real, full");
        }
    }
    
    
    /*
     * Get the manditory inputs.
     */
        
    /* Get the residule 
     *  Performed before approx as the order of inputs changed
     */
    unsigned long nRows = 0;
    unsigned long nColumns = 0;
    double *mResidule = GetResidule(prhs[1],nRows,nColumns);    

    /* Get the current approximation */
    double *mApprox = GetApprox(prhs[0],nRows,nColumns);
    
    /* Get 1D Dictionary selected by MP (horizontal coordinate) performs 
     inner products with the columns of the residule */
    unsigned long nAtomsColumn = 0;
    double *mDictColumns = GetDictColumns(prhs[2],nRows,nAtomsColumn);
    
    /* Get 1D Dictionary selected by MP (vertical coordinate) performs 
     inner products with the rows of the residule */
    double *mDictRows = GetDictRows(prhs[3],nColumns,nAtomsColumn);    
    
    /* Get the coefficientes in the MP approximation */
    double *vCoefficients = GetCoefficients(prhs[4], nAtomsColumn);
    

    /* 
     * Check for the optional inputs
     */
    // <<---
        /* Get the numerical tolerance for the projection (default tre=1e-3) */

        /* MATLAB Code
         *  if (nargin<6) | (isempty(toln)==1)
         *      toln =1e-3;
         *  end
         */
        double tolerance = 0;
        if ((nrhs > 5) && !mxIsEmpty(prhs[5]))
        {
            tolerance = GetTolerance(prhs[5]);
        }else{
            tolerance = 1e-3;
        }

        /* Get the maximum number of iterations (default Max=3000) */

        /* MATLAB Code
         *  if (nargin<7) | (isempty(Max)==1)  
         *      Max=3000;
         *  end
         */
        unsigned long maxIterations = 0;
        if ((nrhs > 6) && !mxIsEmpty(prhs[5]))
        {
            maxIterations = GetMaxIterations(prhs[6]);
        }else{
            maxIterations = 3000;
        }
    // --->>
   
    /* Validate number of output (lhs) arguments*/    
    if((nlhs<3) || (nlhs > 4))
    {
        mexErrMsgTxt("Wrong number of output arguments");
    }
    
    /* Initialize Output variables */
    // <<--
        /* Create mxArrays for the output data */

        /* # of pixels in the residule and approximation */    
        ulong nPixels = nRows*nColumns;    

        /* Create mxArray for projected approximation */
        double *mApproxProjected = SetVar(plhs[0],nRows,nColumns);

        /* Copy accros current approximation */
        CopyElements(mApprox,mApproxProjected,nPixels);

        /* Create mxArray for projected residule */
        double *mResiduleProjected = SetVar(plhs[1],nRows,nColumns);

        /* Copy accros current residule */
        CopyElements(mResidule,mResiduleProjected,nPixels);

        /* Create mxArray for coefficients in the decomposition */
        double *vCoefficientsOut = SetVar(plhs[2],1,nAtomsColumn);

        /* Copy accros current coefficients */
        CopyElements(vCoefficients,vCoefficientsOut,nAtomsColumn);                   
    // -->>
    

    // Extra param to count the number of inner iterations
        ulng iterations = 0;
        
        
    /* Invoke target function */
    ProjMP2D(mResiduleProjected,mApproxProjected,mDictColumns,mDictRows,
            vCoefficientsOut,tolerance,maxIterations,nRows,nColumns,
            nAtomsColumn,iterations);
    

    /* Optional output params */
        if (nlhs ==4)
        {
            /* Initialize optional iteration counter */
            double *iterationsOut = SetVar(plhs[3],1,1);
            *iterationsOut = iterations;
        }    
    

}


double *GetResidule(const mxArray *prhs,unsigned long &nRows,
        unsigned long &nColumns)
{
    /* Get the residule */
    double *mResidule = GetVar(prhs,nRows,nColumns,"Residule");
    
    return mResidule;
    
}

double *GetApprox(const mxArray *prhs, unsigned long nRowsResidule,
        unsigned long nColumnsResidule)
{
    /* Get the approximation */
    unsigned long nRows = 0;
    unsigned long nColumns = 0;
    double *mApprox = GetVar(prhs,nRows,nColumns,"Approx");
    
    /* Validate dimensions */
    ValidateApprox(nRowsResidule,nColumnsResidule,nRows,nColumns);
    
    return mApprox;
    
}

void ValidateApprox(unsigned long nRowsResidule, unsigned 
        long nColumnsResidule, unsigned long nRows, unsigned long nColumns)
{
    /* Validate that nRows and nColumns are consistend with nRowsRedicule
     * and nColumnsResidule respectivly*/
    if ((nRowsResidule != nRows) || (nColumnsResidule != nColumns))
    {
        mexPrintf("mResidule dimensions %i x %i, are not equal to mApprox dimensions %i x %i\n",
                nRowsResidule,nColumnsResidule,nRows,nColumns);
        mexErrMsgTxt("Error!!!!"); 
    }
    
}
    
double *GetDictColumns(const mxArray *prhs, unsigned long nRowsResidule,
        unsigned long &nAtomsColumn)
{
    /* Get the column dictionary */
    unsigned long nRows = 0;
    //unsigned long nColumn = 0;
    double *mDictColumn = GetVar(prhs,nRows,nAtomsColumn,
            "Column Dictionary");    
    
    /* Validate dimensions */
    ValidateDictColumn(nRowsResidule,nRows);
    
    return mDictColumn;
    
}

void ValidateDictColumn(unsigned long nRowsResidule, unsigned long nRows)
{
    /* The column dictionary's atoms performs inner products with the columns of
     * the residule.  Therefore it must have the same number of rows */
    if (nRowsResidule != nRows)
    {
        mexPrintf("Residule has %i rows, column dictionary has %i rows\n",
                nRowsResidule,nRows);
        mexErrMsgTxt("Error column dictionary and redisule have different numbers of rows!!!!");
    }
    
}


double *GetDictRows(const mxArray *prhs,unsigned long nColumnsResidule,
        unsigned long nAtomsColumn)
{
    /* Get the row dictionary */
    unsigned long nRows = 0;
    unsigned long nColumns = 0;
    double *mDictRow = GetVar(prhs,nRows,nColumns,"Row Dictionary");   
        
    /* Validate dimensions */
    ValidateDictRow(nColumnsResidule,nAtomsColumn,nRows,nColumns);
    
    return mDictRow;
    
    
}

void ValidateDictRow(unsigned long nColumnsResidule, unsigned long 
        nAtomsColumn, unsigned long nRows, unsigned long nColumns)
{    
    /* The row dictionary's atoms perform inner products with the rows of the 
     * residule.  Therefore it must have the same number of rows as it
     * has columns.
     */
    if (nColumnsResidule != nRows)
    {
        mexPrintf("Residule has %i columns, row dictionary has %i rows\n",
            nColumnsResidule,nRows);
        mexErrMsgTxt("Error row dictionary and redisule have inconsistent dimensions!!!!"); 
    }
    
    /* The row dictionary must also have the same number of atoms as the 
     * column dictionary.
     */
    if (nColumns != nAtomsColumn)
    {
        mexPrintf("Row dictionary has %i atoms, Column dictionary has %i atoms\n",
            nColumns, nAtomsColumn);
        mexErrMsgTxt("Number of atoms should be the same!!");
    }
   
}

double *GetCoefficients(const mxArray *prhs, unsigned long nAtoms)
{
    /* Get the coefficient vector*/
    unsigned long nRows = 0;
    unsigned long nColumns = 0;
    double *vCoefficients = GetVar(prhs,nRows,nColumns,"Coefficients");  
    
    ValidateCoefficients(nAtoms,nRows,nColumns);
    
    return vCoefficients;
    
}

void ValidateCoefficients(unsigned long nAtoms, unsigned long nRows,
        unsigned long nColumns)
{
    /* Checks that the number of coefficients is equal to the number of 
     * atoms */
    
    /* The largest dimension is the number of coefficients */
    unsigned long nCoefficients = nColumns;
    if (nRows > nColumns)
    {
        nCoefficients = nRows;
    }
    
    /* Validate */
    if (nCoefficients != nAtoms)
    {
        mexPrintf("# of atoms %i, # of coefficients %i\n",
            nAtoms, nCoefficients);
        mexErrMsgTxt("Number of coefficients should be equal to the number of atoms!!");                
    }
    
}

double GetTolerance(const mxArray *prhs)
{
    /* Get the tolerance*/
    unsigned long nRows = 0;
    unsigned long nColumns = 0;
   // mexPrintf("Before tol cast\n");
    double tolerance = *GetVar(prhs,nRows,nColumns,"Tolerance");
   // mexPrintf("After tol cast\n");
    /* Validate scalar */
    ValidateScalar(nRows,nColumns,"Tolerance");
    
    return tolerance;
    
}

    
unsigned long GetMaxIterations(const mxArray *prhs)
{
    /* Get the maximum # of iterations*/
    unsigned long nRows = 0;
    unsigned long nColumns = 0;
    unsigned long maxIterations = static_cast<unsigned long>
        (*GetVar(prhs,nRows,nColumns,"Max Iterations"));

    /* Validate scalar */
    ValidateScalar(nRows,nColumns,"Max Iterations");
    
    return maxIterations; 
    
}


void ValidateScalar(unsigned long nRows, unsigned long nColumns, const char 
        *varName)
{
    
    if ((nRows != 1) || (nColumns != 1))
    {
        mexPrintf("%s has %i rows and %i columns\n", varName, nRows, 
                nColumns);
        mexErrMsgTxt("Variable should be a scalar!!");
    }
    
}

double *GetVar(const mxArray *prhs, unsigned long &M,
        unsigned long &N, const char *varName)
{
    /* Gets a pointer to the double array stored in the mxArray prhs[0] 
     * and retrieves the dimensions of the array.
     *  Validates that dimensions are not empty (0).
     */
    
    /* Get pointer to array */
    double *var = mxGetPr(prhs);
    
    /* Get array dimensions */
    M = mxGetM(prhs);
    N = mxGetN(prhs);
    
    /* Validate that array is not empty */
    if ((M == 0) || (N == 0))
    {
        mexPrintf("Input array %s has zero dimensions\n",varName);
        mexErrMsgTxt("Error!!!!");
    }
    
    return var;

}

double *SetVar(mxArray *&m, unsigned long nRows, unsigned long 
        nColumns) // Use reference to pointer
{
    /* Creates a real mxArray, size nRows x nColumns assigns it to m[0]
     * and returns the 
     * pointer to it */
    
   m = mxCreateDoubleMatrix(nRows, nColumns, mxREAL);
    
  //  double *t = mxGetPr(m);
   // t[0] = 1;
    return mxGetPr(m);
}



