function [H,Di1,Di2,beta,c,Q,Nat]=OHBW_OMP2D(f,Dx,Dy,tol,No);
%Optimized Hierarchical Block Wise Orthogonal Matching Pursuit in 2D
% 
%Separable dictionaries version OHBW_OMP2D is the working version
% 2/1/2014
% It creates an atomic decomposition of a 2D signal divided into blocks and  assuming 
% separable dictionary.  At each iteration the algorithms selects
% i)the elements for approximating  the blocks in the partition, using OMP criterion
% ii)the hierarchical sequence in which the blocks are approximated.
% You can choose a tolerance for the approximation or/and the total number of atoms to take in. 
% 
%  
% Usage:    [H,Di1,Di2, beta, c,Q,Nat] = HBW_OMP2D(f,Dx,Dy, tol, No);
%           [H] = OMP2D(f,Dx,Dy); variables tol, No, can also be []
%                 
%
% Inputs:
%   f       cell  array f{n,m} is a matrix with the pixels of block n,m of the image
%   Dx      dictionary of normalized atoms (w.r.t. Image rows)
%   Dy      dictionary of normalized atoms (w.r.t. Image columns)
%   tol     desired distance between f and its approximation H  
%   No      maximal number of atoms to choose, if the number of chosen atoms
%           equals to No, routine will stop (default numel(f))
%
% Outputs:
%   H       cell array with the approximation of f
%   Di1     cell with the indices of selected atoms for the block w.r.t the original Dx
%   Di2     cell  with the indices of selected atoms for the block w.r.t thew.r.t the original Dy
%   beta    cell array with biorthogonal atoms corresponding to the 2D selected atoms 
%   c       cell array with coefficients of the atomic decomposition
%   Q       cell array with an ortho basis for the selected subspace
%   Nat     Total number of atoms (in case it is not No)
%  
%   
% The implementation of OMP2D method is based on Gram-Schmidt orthonormalization and 
% adaptive birothogonalization, as proposed in Ref 2 below.
%
%----------------------------------------------------------
%  Laura Rebollo-Neira
%  http://www1.aston.ac.uk/eas/staff/dr-laura-rebollo-neira/
%----------------------------------------------------------
%
%  Reference:
%
%  1-``Hierarchical Block Wise Image Approximation by Greedy Pursuit Strategies.''
%  
%  2-L. Rebollo-Neira and D. Lowe, "Optimized Orthogonal Matching Pursuit Approach", IEEE
%  Signal Processing Letters, Vol(9,4), 137-140, (2002). 
%
% See also OMP2D OMP2DMl OMPMl OOMPMl OMP2D_Block and all Pursuit routines on 
%
% http://www.nonlinear-approx.info/

name='OHBW-OMP2D'; %name of routine
% get inputs and setup parameters
%!!!! modification of OMP 1D
[Lx,Nx]=size(Dx);
[Ly,Ny]=size(Dy);
%delta=1/(Lx*Ly);
delta=1;
NBx=size(f,1);
NBy=size(f,2);
N=NBx*NBy;
zmax=1;%number of reorthogonalizations
beta=cell(N,1);
Q=cell(N,1);
Q2=cell(N,1);
c=cell(N,1);
cc=cell(N,1);
qq=cell(N,1);
Di1=cell(N,1);
Di2=cell(N,1);
Re=f;
nn=0;
%for ii=1:NBx
%for jj=1:NBy
%nn=nn+1;
%B=Re{ii,jj};
%nore(nn)=norm(B(:))^2;
%end
%end

kq=zeros(N,1);
Cp=zeros(N,1);
Cpsi=zeros(N,1);
if (nargin<5) | (isempty(No)==1)  No=9*N;end
if (nargin<4) | (isempty(tol)==1) tol=0.03*sum(nore);end; 
%
%atoms having smaller norm than tol1 are supposed be zero ones
tol1=1e-7; %1e-5
%threshold for coefficients
tol2=1e-11;   
Nat=0;
nb=0;
for i=1:NBx
for j=1:NBy
    nb=nb+1;
    iq(nb)=i;
    jq(nb)=j;
    B=Re{i,j};
    nore(nb)=norm(B(:))^2;
    cc{nb}=abs(Dx'*B*Dy);
   [c1,c2]=max(cc{nb});
   [max_c,b2]=max(c1);
    qq{nb}=[c2(b2) b2];
    q(2)= qq{nb}(2);
    q(1)= qq{nb}(1);
    Cp(nb)=max_c;
    Cpsi(nb)=Cp(nb);
    Q{nb}=kron(Dy(:,q(2)),Dx(:,q(1)));
    norb=norm(Q{nb});
    Q{nb}=Q{nb}/norb;
    beta{nb}=Q{nb}/norb;
end
end
%
%===============================
% Main algorithm 
%===============================
while Nat < No 
  %finding of maximal coefficient
    [Cpmax,qb]=max(Cpsi);
    q(1)=qq{qb}(1);
    q(2)=qq{qb}(2);
  %stopping criterion (coefficient)
    if Cpmax<tol2 
      fprintf('%s stopped, max(|<f,q>|/||q||)<= tol2=%g.\n',name,tol2);
      break;
    end
    kq(qb)=kq(qb)+1;
    k=kq(qb);
    Di1{qb}=[Di1{qb} q(1)];
    Di2{qb}=[Di2{qb} q(2)];
  if k>1
  new_atom=kron(Dy(:,q(2)),Dx(:,q(1)));
  nork=norm(Q{qb}(:,k));
  Q{qb}(:,k)=Q{qb}(:,k)/nork;
  beta{qb}=beta{qb}-Q{qb}(:,k)*(new_atom'*beta{qb})/nork;
  beta{qb}(:,k)=Q{qb}(:,k)/nork;
  end
  h=f{iq(qb),jq(qb)}(:)'*Q{qb}(:,k)*Q{qb}(:,k)';
  Re{iq(qb),jq(qb)}=Re{iq(qb),jq(qb)}(:)-h';%residual of block qp
  nore(qb)=(norm(Re{iq(qb),jq(qb)}(:)))^2*(delta); %residual norm of block qp
  Re{iq(qb),jq(qb)}=reshape(Re{iq(qb),jq(qb)},Lx,Ly);
  c{qb}=f{iq(qb),jq(qb)}(:)'*beta{qb};
  cc{qb}=abs(Dx'*Re{iq(qb),jq(qb)}*Dy);
  [c1,c2]=max(cc{qb});
  [max_c,b2]=max(c1);
  qq{qb}=[c2(b2) b2];
  q(1)=qq{qb}(1);
  q(2)=qq{qb}(2);
  Cp(qb)=max_c;

%=====adding the Psi thing 
  new_atom=kron(Dy(:,q(2)),Dx(:,q(1)));
  for p=1:k
     QQ=new_atom-(Q{qb}(:,p)'*new_atom)*Q{qb}(:,p); %orthogonalization
  end
   for zi=1:zmax
    for p=1:k
     QQ=QQ-(Q{qb}(:,p)'*QQ)*Q{qb}(:,p); %re-orthogonalization
    end
   end
  Q{qb}(:,k+1)=QQ;
  QQ=QQ/norm(QQ);
  Cpsi(qb)=abs(QQ'*Re{iq(qb),jq(qb)}(:));
%=====================================================
% 
  Nat=Nat+1;
  max_err=sum(nore);
  %stopping criterion (distance)
  if (tol~= 0) & (max_err < tol) break;end;
end
max_err
%!!!approximation and selected atoms
H=cell(NBx,NBy);
nq=0;
for i=1:NBx
for j=1:NBy
BH=zeros(Lx,Ly);
nq=nq+1;
H{i,j}=zeros(Lx,Ly);
id=numel(c{nq});
for n=1:id;
BH=BH+Dx(:,Di1{nq}(n))*c{nq}(n)*Dy(:,Di2{nq}(n))';
end
H{i,j}=BH;
end
end
%
%
%Copyright (C) 2011  Laura REBOLLO-NEIRA, Mathematics Department, Aston University
%
%This program is free software; you can redistribute it and/or modify it under the terms 
%of the GNU General Public License as published by the Free Software Foundation; either 
%version 2 of the License, or (at your option) any later version.
%
%This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
%without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
%See the GNU General Public License for more details.
%
%You should have received a copy of the GNU General Public License along with this program;
%if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
%Boston, MA  02110-1301, USA.
